# MoleskineApp.py
#
# Moleskine: a source code editor for the GNOME desktop
#
# Copyright (c) 2001   Michele Campeotto <micampe@f2s.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Library General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

import os.path

import gtk
import GTK
import GDK
import gnome.ui
import gnome.uiconsts

from Document import Document
from Config import Config
from Languages import LanguagesList
import menu
import toolbar

class MoleskineApp(gnome.ui.GnomeApp):
    def __init__(self):
        gnome.ui.GnomeApp.__init__(self, 'GnomeMoleskine', 'Moleskine')
        
        self.switch = 1
        self.documents = []
        self.load_prefs()
        
        self.notebook = gtk.GtkNotebook()
        self.notebook.connect_after('switch-page', self.__switch_page)
        self.notebook.set_scrollable(1)
        self.notebook.show()
        
        self.set_contents(self.notebook)
        self.set_wmclass('main_window', 'Moleskine')
        self.set_geometry_hints(min_width=0, min_height=0)
        self.connect('key_press_event', self.key_press)
        if int(self.prefs['keyboard/keybindings']) == 2:
            import Emacs
            self.create_menus(Emacs.main_menu)
            self.Emacs_key_press = Emacs.key_press
            #self.connect('key_press_event', Emacs.key_press)
        elif int(self.prefs['keyboard/keybindings']) == 1:
            import GnomeMenu
            self.create_menus(GnomeMenu.main_menu)
        else:
            self.create_menus(menu.main_menu)
        self.create_toolbar(toolbar.main_toolbar)
        self.connect('delete_event', menu.exit)
        self.show_all()
        
        self.load_langs()
        self.load_mru()
    
    def key_press(self, w, event):
        """Handle keypresses."""
        if event.state & GDK.MOD1_MASK and not event.state & (GDK.SHIFT_MASK | GDK.CONTROL_MASK):
            if event.keyval >= GDK._1 and event.keyval <= GDK._9:
                self.notebook.set_page(event.keyval - GDK._1)
            elif event.keyval == GDK._0:
                self.notebook.set_page(9)
        else:
            if int(self.prefs['keyboard/keybindings']) == 2:
                self.Emacs_key_press(w, event)

    def new_document(self):
        """Create a new document."""
        doc = Document()
        self.documents.append(doc)
        self.notebook.append_page(doc, doc.label)
        self.notebook.set_page(-1)
        self.update_title()
    
    def open_document(self, filepath):
        """Open a document.
        
        The file name can contain relative paths, special shell chars (~) and
        environment variables.
        If the current document is empty the newly created one will replace it.
        If the given file doesn't exist you will be asked about to create it.
        """
        
        filepath = os.path.expanduser(filepath)
        filepath = os.path.expandvars(filepath)
        filepath = os.path.abspath(filepath)
        
        if os.path.isdir(filepath):
            gnome.ui.GnomeMessageBox(
                'Cannot open `%s\': is a directory' % \
                os.path.basename(filepath),
                gnome.uiconsts.MESSAGE_BOX_ERROR,
                gnome.uiconsts.STOCK_BUTTON_OK).run_and_close()
        elif not os.path.exists(filepath):
            res = gnome.ui.GnomeMessageBox(
                    'File `%s\' not found.\nDo you want to create it?' % \
                            os.path.basename(filepath),
                    gnome.uiconsts.MESSAGE_BOX_QUESTION,
                    gnome.uiconsts.STOCK_BUTTON_YES,
                    gnome.uiconsts.STOCK_BUTTON_NO).run_and_close()
            if res == 0:
                self.__load(filepath)
                self.get_current_document().set_text(' ')
                self.get_current_document().set_text('')
            else:
                if len(self.documents) == 0:
                    self.new_document()
        elif not os.access(filepath, os.R_OK):
            gnome.ui.GnomeMessageBox(
                'Cannot open `%s\': permission denied.' % \
                os.path.basename(filepath),
                gnome.uiconsts.MESSAGE_BOX_ERROR,
                gnome.uiconsts.STOCK_BUTTON_OK).run_and_close()
            return
        else:
            self.__load(filepath)
        self.get_current_document().grab_focus()
        self.update_title()
    
    def __load(self, filepath):
        doc = self.get_current_document()
        
        for i in range(len(self.documents)):
            if self.documents[i].filepath == filepath:
                self.notebook.set_page(i)
                if not self.documents[i].is_modified():
                    self.documents[i].revert()
                return
        if doc is None or not doc.is_empty():
            self.new_document()
            doc = self.get_current_document()
        doc.load_file(filepath)
        self.add_to_mru(filepath)
    
    def save_document(self, filepath = None):
        self.get_current_document().save_file(filepath)
    
    def close_document(self):
        """Close the active document.
        
        If the document to be closed is the last one, a new empy document will
        be created.
        If the document has been modified and not saved you will be asked if
        you want to save the changes.
        """
        doc = self.get_current_document()
        if not doc.can_close():
            res = gnome.ui.GnomeMessageBox(
                    'File `%s\' has been modified\nand not saved.'
                    ' Do you want to save it now?' % doc.filename,
                    gnome.uiconsts.MESSAGE_BOX_QUESTION,
                    gnome.uiconsts.STOCK_BUTTON_YES, 
                    gnome.uiconsts.STOCK_BUTTON_NO,
                    gnome.uiconsts.STOCK_BUTTON_CANCEL).run_and_close()
            if res == 0:
                doc.save_file()
            elif res == 2 or res < 0:
                return
        page = self.notebook.get_current_page()
        self.notebook.remove_page(page)
        self.documents.remove(doc)
        if len(self.documents) == 0:
            self.new_document()
        else:
            self.notebook.set_page(page)
    
    def ask_file(self, type, ok_cb):
        if type == 'load':
            title = 'Open document...'
            wm_name = 'open_dialog'
        elif type == 'save':
            title = 'Save as...'
            wm_name = 'save_dialog'
        elif type == 'load_session':
            title = 'Open session...'
            wm_name = 'open_dialog'
        elif type == 'save_session':
            title = 'Save session as...'
            wm_name = 'save_dialog'
        else:
            title = type
            wm_name = 'file_dialog'
        
        fs = gtk.GtkFileSelection(title)
        #fs.set_wmclass(wm_name, 'Moleskine')
        fs.set_transient_for(self)
        fs.file_list.set_selection_mode(GTK.SELECTION_MULTIPLE)
        
        doc = self.get_current_document()
        if doc is not None and doc.filepath is not None:
            startdir = os.path.dirname(doc.filepath)
            fs.set_filename(startdir + '/')
        if ok_cb is not None:
            fs.ok_button.connect_object('clicked', ok_cb, fs)
        
        fs.ok_button.connect_object('clicked', gtk.GtkWidget.destroy, fs)
        fs.cancel_button.connect_object('clicked', gtk.GtkWidget.destroy, fs)
        fs.show()
    
    def get_current_document(self):
        """Returns the currently active document.
        
        If there are no documents (should never happen) it will return None.
        """
        if len(self.documents) > 0:
            return self.documents[self.notebook.get_current_page()]
        else:
            return None
    
    def load_mru(self):
        self.mru = Config('Moleskine/PreviousFiles')
        self.files = []
        count = self.mru['files/count']
        if count is not None:
            for i in range(int(count)):
                self.files.append(self.mru['files/file_%i' % i])
            self.update_mru_menu()
    
    def save_mru(self):
        self.files = self.files[:int(self.prefs['history/mru_size'])]
        for i in range(len(self.files)):
            self.mru['files/file_%i' % i] = self.files[i]
        self.mru['files/count'] = len(self.files)
        self.mru.save()
    
    def update_mru_menu(self):
        import copy, string
        
        fix_underscore = lambda label, replace=string.replace: \
                replace(label, '_', '__')
        
        build_item = lambda label: \
                (gnome.uiconsts.APP_UI_ITEM, label, None,
                 menu.open_recent, None,
			     gnome.uiconsts.APP_PIXMAP_NONE, None, 0, 0)
        
        temp_list = copy.copy(self.files)[:int(self.prefs['history/mru_size'])]
        if int(self.prefs['history/mru_sort']):
            temp_list.sort()
        temp_list = map(fix_underscore, temp_list)
        mru_menu = map(build_item, temp_list)
        self.remove_menus('File/Open previous/', 55)
        self.insert_menus('File/Open previous/', mru_menu)
    
    def add_to_mru(self, filepath):
        if filepath in self.files:
            self.files.remove(filepath)
        self.files.insert(0, filepath)
        if len(self.files) > int(self.prefs['history/mru_size']):
            self.files = self.files[:int(self.prefs['history/mru_size'])]
        self.update_mru_menu()
    
    def load_session(self, filename = None):
        if filename is not None:
            self.session = Config('=' + filename + '=')
        else:
            self.session = Config('Moleskine/Session')
        count = self.session['files/count']
        if count is not None and int(count) > 0:
            if filename is not None:
                self.notebook.set_page(0)
                for i in range(len(self.documents)):
                    self.close_document()
            self.switch = 0
            document_index = 0
            for i in range(int(count)):
                if self.session['files/file_%i' % i] == 'None':
                    self.new_document()
                    document_index = document_index + 1
                else:
                    self.open_document(self.session['files/file_%i' % i])
                    if (len(self.documents) - document_index) == 1:
                        if self.session['langs/lang_%i' % i] in self.langs.keys():
                            self.documents[document_index].set_language(self.langs[self.session['langs/lang_%i' % i]])
                        if self.session['pos/pos_%i' % i] is not None:
                            self.documents[document_index].goto_pos(int(self.session['pos/pos_%i' % i]))
                        document_index = document_index + 1
            self.switch = 1
            if not document_index:
                self.new_document()
            self.notebook.set_page(int(self.session['notebook/tab']))
            self.update_title()
    
    def save_session(self, filename = None):
        if int(self.prefs['history/save_session']) or filename is not None:
            if filename is not None:
                self.session = Config('=' + filename + '=')
            else:
                self.session = Config('Moleskine/Session')
            self.session['files/count'] = len(self.documents)
            for i in range(len(self.documents)):
                self.session['files/file_%i' % i] = self.documents[i].filepath
                self.session['langs/lang_%i' % i] = self.documents[i].language.name
                self.session['pos/pos_%i' % i] = self.documents[i].get_current_pos()
            self.session['notebook/tab'] = self.notebook.get_current_page()
            self.session.save()
        else:
            self.session = Config('Moleskine/Session')
            self.session.data.clear()
            self.session['files/count'] = 0
            self.session.save()
    
    def load_prefs(self):
        self.prefs = Config('Moleskine/Preferences')
        
        self.prefs.dirnames = []
        count = int(self.prefs['directory names/count'])
        for n in range(count):
            name = self.prefs['directory names/name_%i' % n]
            directory = self.prefs['directory names/directory_%i' % n]
            
            directory = os.path.expanduser(directory)
            directory = os.path.expandvars(directory)
            directory = os.path.abspath(directory)
            
            self.prefs.dirnames.append((name, directory))
        
        w = int(self.prefs['geometry/width'])
        h = int(self.prefs['geometry/height'])
        self.set_default_size(w, h)
        
        self.prefs['view/ws'] = 0
        self.prefs['view/eol'] = 0
    
    def save_prefs(self):
        count = len(self.prefs.dirnames)
        self.prefs['directory names/count'] = count
        for n in range(count):
            self.prefs['directory names/name_%i' % n] = \
                    self.prefs.dirnames[n][0]
            self.prefs['directory names/directory_%i' % n] = \
                    self.prefs.dirnames[n][1]
        
        if int(self.prefs['geometry/save_win_size']):
            self.prefs['geometry/width'] = self.get_window().width
            self.prefs['geometry/height'] = self.get_window().height
    
    # FIXME: Can't get the right window position! :( This gives
    # the position of the client window (without decorations)
    # and set_uposition places the window with deocorations. Hints?
    #    if int(self.prefs['geometry/save_win_pos']):
    #        self.prefs['geometry/pos_x'] = self.get_window().x
    #        self.prefs['geometry/pos_y'] = self.get_window().y
        self.prefs.save()
    
    def load_langs(self):
        self.langs = LanguagesList()
        self.langs.load()
        
        langs_names = self.langs.keys()
        langs_names.sort()
        build_item = lambda label: \
                (gnome.uiconsts.APP_UI_ITEM, label, None,
                 menu.set_lang, None,
			     gnome.uiconsts.APP_PIXMAP_NONE, None, 0, 0)
        langs_menu = [gnome.uiconsts.UIINFO_RADIOLIST(
                          map(build_item, langs_names))]
        self.remove_menus('Settings/Change language/', len(langs_names))
        self.insert_menus('Settings/Change language/', langs_menu)
    
    def update_styles(self):
        for doc in self.documents:
            doc.setup_default_style()
            doc.setup_caret_style()
            doc.setup_line_numbers_style()
            doc.reload_language()
    
    def update_title(self):
        if not self.switch:
            return
        doc = self.get_current_document()
        if doc.is_modified():
            sep = ' * '
        else:
            sep = ' - '
        self.set_title('Moleskine' + sep + doc.title)
    
    def __switch_page(self, notebook, notebook_page, page_num, data=None):
        if not self.switch:
            return
        doc = self.get_current_document()
        self.update_title()
        self.get_current_document().set_view_ws(self.prefs['view/ws'])
        self.get_current_document().set_view_eol(self.prefs['view/eol'])
        try:
            langs_names = self.langs.keys()
            langs_names.sort()
            index = langs_names.index(doc.language.name)
            menu_items = self.langs_menu.children()
            if len(langs_names) < len(menu_items):
                menu_items = menu_items[1:]
            menu_items[index].set_active(1)
        except AttributeError:
            pass
        doc.grab_focus()
